/* 
 * Generic XML Configuration Access Library
 * 
 * Copyright (C) 2005 Robert Schwebel <r.schwebel@pengutronix.de>
 *		
 */

#include <ctype.h>
#include <errno.h>
#include <string.h>
#include <stdlib.h>
#include <stdio.h>
#include <libxml/parser.h>
#include <libxml/xpath.h>

#include <xmlconfig.h>

static const char prompt[] = "libxmlconfig";

enum xmlconfig_convert_type {
	CONV_FLOAT,
	CONV_INT,
	CONV_UINT,
	CONV_STRING,
};

#ifndef MIN
#define MIN(a,b) (((a)<(b))?(a):(b))
#endif

/**
 * xmlconfig_init:
 * @filename: name of an XML configuration file
 *
 * This is the constructor function for xmlconfig_t objects. It opens
 * and parses an XML config file. 
 *
 * Return value: On success a handle to the xmlconfig_t object is
 * returned. In case of an error this function returns NULL. 
 **/
xmlconfig_t *xmlconfig_init(const char *filename)
{
	xmlconfig_t *cfg; 

	/* allocate */
	cfg = calloc(1, sizeof(xmlconfig_t));
	if (!cfg) {
		fprintf(stderr, "%s: error: out of memory\n", prompt);
		return NULL; 
	}

	/* parse */
	cfg->doc = xmlParseFile((const xmlChar*)filename);
	if (!cfg->doc) {
		fprintf(stderr, "%s: could not open %s\n", prompt, filename);
		free(cfg);
		return NULL;
	}
	
	return cfg;
}


/**
 * xmlconfig_finish:
 * @cfg: handle to an xmlconfig object
 *
 * Destructor for xmlconfig_t objects. This function frees an object,
 * which means that all memory allocated by xmlconfig_init() is given
 * back. 
 *
 * Return value: This function returns 0 on success. 
 **/
int xmlconfig_finish(xmlconfig_t *cfg)
{
	xmlFreeDoc(cfg->doc);
	xmlCleanupParser();
	free(cfg); 
	xmlMemoryDump();
	return 0; // FIXME: not always success
}


static int xmlconfig_get (
	xmlconfig_t *cfg, 
	enum xmlconfig_convert_type type, 
	int strlength, 
	char *xpath, 
	void *val, 
	int nelem
)
{
	int numfound = 0;
	int i;
	xmlXPathContextPtr context;
	xmlChar *content;

	char *endptr;
	char *startptr;
	
	unsigned long *ulval = 0;
	long *lval = 0;
	double *dval = 0; 
	char *cval = NULL; 

	if (!cfg)
		return -EINVAL;

	if (!val)
		return -EINVAL;

	switch(type) {
		case CONV_UINT:
			ulval = (unsigned long*)val;
			break;
		case CONV_INT:
			lval = (long*)val;
			break;
		case CONV_FLOAT:
			dval = (double*)val;
			break;
		case CONV_STRING:
			cval = (char*)val;
			break;
		default:
			return -EINVAL;
	}

	context = xmlXPathNewContext(cfg->doc);
	if (context == NULL) {
		fprintf(stderr, "%s: couldn't get xpath context\n", prompt);
		return -ENOMEM;
	}

	cfg->result = xmlXPathEvalExpression((xmlChar*)xpath, context);
	xmlXPathFreeContext(context);
	if (!cfg->result) {
		fprintf(stderr, "%s: error in xmlXPathEvalExpression\n", prompt);
		return -EFAULT;
	}

	numfound = (cfg->result->nodesetval) ? cfg->result->nodesetval->nodeNr : 0; 

	for (i=0; i<numfound; i++) {

		if (i>=nelem) break;
		
		content = xmlNodeGetContent(cfg->result->nodesetval->nodeTab[i]);
		startptr = content;
		while (isspace(*startptr)){startptr++;}
		if ((type == CONV_UINT) && (*startptr == '-')) {
			xmlFree(content);
			return -ERANGE;
		}
		
		switch (type) {
			case CONV_UINT: 
				*ulval = strtoul((char*)content, &endptr, 0); 
				break;
			case CONV_INT: 
				*lval = strtol((char*)content, &endptr, 0); 
				break;
			case CONV_FLOAT:
				*dval = strtod((char*)content, &endptr); 
				break;
			case CONV_STRING:
				/* FIXME: can a char[20] have 20 chars _plus_ \0? */
				strncpy(cval,content,MIN(strlen(content),strlength-1)); 
				cval[MIN(strlen(content),strlength-1)]='\0';
				break;
		}
		
		if (type != CONV_STRING) {
			while (isspace(*endptr)){endptr++;}
			if ((content[0] == '\0') || (*endptr != '\0')) {
				xmlFree(content);
				return -ERANGE;
			}
		}

		xmlFree(content);
		
		switch (type) {
			case CONV_UINT:
				ulval++; 
				break;
			case CONV_INT:
				lval++; 
				break;
			case CONV_FLOAT:
				dval++;
				break;
			case CONV_STRING:
				cval+=strlength;
				break;
		}
	}

	xmlXPathFreeObject(cfg->result);
	return numfound; 
}


/**
 * xmlconfig_get_ulong
 * @cfg: handle to an xmlconfig object
 * @xpath: XPath expression which describes an element set
 * @val: pointer to returned unsigned long value
 * @nelem: number of elements to be found
 *
 * This function parses the XML file for a tag matching the XPath expression 
 * given in @xpath. The calling code can specify the number of expected hits 
 * in @nelem. In case of success the resulting content is written to an array
 * of size @nelem located at @val. If nelem is 0 no value is given back
 * but the correct number of hits is returned. 
 *
 * Return value: Values >= 0 indicate the number of matches which have
 * been found. Negative return values show an error: -ENOMEM: some internal 
 * allocation failed. -EINVAL: incorrect function parameters. 
 * -EFAULT: Malformed XPath expression. -ERANGE: match not convertable
 * to unsigned long. In case of an error *val doesn't necessarily
 * contain useful content. 
 **/
int xmlconfig_get_ulong (xmlconfig_t *cfg, char *xpath, unsigned long *val, int nelem)
{
	return xmlconfig_get(cfg, CONV_UINT, 0, xpath, (void*)val, nelem);
}

/**
 * xmlconfig_get_ulong
 * @cfg: handle to an xmlconfig object
 * @xpath: XPath expression which describes an element set
 * @val: pointer to returned unsigned long value
 * @nelem: number of elements to be found
 *
 * This function parses the XML file for a tag matching the XPath expression 
 * given in @xpath. The calling code can specify the number of expected hits 
 * in @nelem. In case of success the resulting content is written to an array
 * of size @nelem located at @val. If nelem is 0 no value is given back
 * but the correct number of hits is returned. 
 *
 * Return value: Values >= 0 indicate the number of matches which have
 * been found. Negative return values show an error: -ENOMEM: some internal 
 * allocation failed. -EINVAL: incorrect function parameters. 
 * -EFAULT: Malformed XPath expression. -ERANGE: match not convertable
 * to unsigned long. In case of an error *val doesn't necessarily
 * contain useful content. 
 **/
int xmlconfig_get_long (xmlconfig_t *cfg, char *xpath, long *val, int nelem)
{
	return xmlconfig_get(cfg, CONV_INT, 0, xpath, (void*)val, nelem);
}

/**
 * xmlconfig_get_double
 * @cfg: handle to an xmlconfig object
 * @xpath: XPath expression which describes an element set
 * @val: pointer to returned double value
 * @nelem: number of elements to be found
 *
 * This function parses the XML file for a tag matching the XPath expression 
 * given in @xpath. The calling code can specify the number of expected hits 
 * in @nelem. In case of success the resulting content is written to an array
 * of size @nelem located at @val. If nelem is 0 no value is given back
 * but the correct number of hits is returned. 
 *
 * Return value: Values >= 0 indicate the number of matches which have
 * been found. Negative return values show an error: -ENOMEM: some internal 
 * allocation failed. -EINVAL: incorrect function parameters. 
 * -EFAULT: Malformed XPath expression. -ERANGE: match not convertable
 * to double. In case of an error *val doesn't necessarily
 * contain useful content. 
 **/
int xmlconfig_get_double (xmlconfig_t *cfg, char *xpath, double *val, int nelem)
{
	return xmlconfig_get(cfg, CONV_FLOAT, 0, xpath, (void*)val, nelem);
}


/**
 * xmlconfig_get_string
 * @cfg: handle to an xmlconfig object
 * @xpath: XPath expression which describes an element set
 * @buf: pointer to returned string
 * @strlength: length of one string
 * @nelem: number of elements to be found
 *
 * This function parses the XML file for a tag matching the XPath expression 
 * given in @xpath. The calling code can specify the number of expected hits 
 * in @nelem. In case of success the resulting content is written to an array
 * of size @nelem located at @val; . If nelem is 0 no value is given back
 * but the correct number of hits is returned. 
 *
 * Return value: Values >= 0 indicate the number of matches which have
 * been found. Negative return values show an error: -ENOMEM: some internal 
 * allocation failed. -EINVAL: incorrect function parameters. 
 * -EFAULT: Malformed XPath expression. In case of an error *val doesn't
 * necessarily contain useful content. 
 **/
int xmlconfig_get_string (xmlconfig_t *cfg, char *xpath, char *buf, int strlength, int nelem)
{
	return xmlconfig_get(cfg, CONV_STRING, strlength, xpath, (void*)buf, nelem);
}


/////////////////////////////////////////////////////////////////////////////////

#if 0
xmlXPathObjectPtr getnodeset (xmlDocPtr doc, xmlChar *xpath){
	
}
#endif

